@@ -386,15 +386,37 @@ escc_add_channel(const char *path, const char *node, phys_addr_t addr,
     phandle_t dnode, aliases;
 
     cell props[10];
-    int offset;
+    ucell offset;
+    int index;
     int legacy;
-
+    
+    int dbdma_offsets[2][2] = {
+        /* ch-b */
+        { 0x6, 0x7 },
+        /* ch-a */
+        { 0x4, 0x5 }
+    };
+    
+    int reg_offsets[2][2][3] = {
+        {
+            /* ch-b */
+            { 0x00, 0x10, 0x40 },
+            /* ch-a */
+            { 0x20, 0x30, 0x50 }
+        },{
+            /* legacy ch-b */
+            { 0x0, 0x2, 0x8 },
+            /* legacy ch-a */
+            { 0x4, 0x6, 0xa }
+        }
+    };
+    
     switch (esnum) {
-    case 2: offset = 1; legacy = 0; break;
-    case 3: offset = 0; legacy = 0; break;
-    case 4: offset = 1; legacy = 1; break;
-    case 5: offset = 0; legacy = 1; break;
-    default: return;
+        case 2: index = 1; legacy = 0; break;
+        case 3: index = 0; legacy = 0; break;
+        case 4: index = 1; legacy = 1; break;
+        case 5: index = 0; legacy = 1; break;
+        default: return;
     }
 
     /* add device */
@@ -425,48 +447,46 @@ escc_add_channel(const char *path, const char *node, phys_addr_t addr,
     set_property(dnode, "compatible", buf, 9);
 
     if (legacy) {
-        props[0] = IO_ESCC_LEGACY_OFFSET + offset * 0x4;
-        props[1] = 0x00000001;
-        props[2] = IO_ESCC_LEGACY_OFFSET + offset * 0x4 + 2;
-        props[3] = 0x00000001;
-        props[4] = IO_ESCC_LEGACY_OFFSET + offset * 0x4 + 6;
-        props[5] = 0x00000001;
-        set_property(dnode, "reg", (char *)&props, 6 * sizeof(cell));
+        offset = IO_ESCC_LEGACY_OFFSET;
     } else {
-        props[0] = IO_ESCC_OFFSET + offset * 0x20;
-        props[1] = 0x00000020;
-        set_property(dnode, "reg", (char *)&props, 2 * sizeof(cell));
+        offset = IO_ESCC_OFFSET;
     }
 
-    if (legacy) {
-        props[0] = addr + IO_ESCC_LEGACY_OFFSET + offset * 0x4;
-    } else {
-        props[0] = addr + IO_ESCC_OFFSET + offset * 0x20;
-    }
+    props[0] = offset + reg_offsets[legacy][index][0];
+    props[1] = 0x1;
+    props[2] = offset + reg_offsets[legacy][index][1];
+    props[3] = 0x1;
+    props[4] = offset + reg_offsets[legacy][index][2];
+    props[5] = 0x1;
+    props[6] = 0x8000 + dbdma_offsets[index][0] * 0x100;
+    props[7] = 0x100;
+    props[8] = 0x8000 + dbdma_offsets[index][1] * 0x100;
+    props[9] = 0x100;
+    set_property(dnode, "reg", (char *)&props, 10 * sizeof(cell));
+
+    props[0] = addr + offset + reg_offsets[legacy][index][0];
     OLDWORLD(set_property(dnode, "AAPL,address",
             (char *)&props, 1 * sizeof(cell)));
 
-    props[0] = 0x00000010 - offset;
+    props[0] = 0x10 - index;
     OLDWORLD(set_property(dnode, "AAPL,interrupts",
             (char *)&props, 1 * sizeof(cell)));
 
-    props[0] = (0x24) + offset;
-    props[1] = 0;
-    props[2] = 0;
+    props[0] = (0x24) + index;
+    props[1] = 0x1;
+    props[2] = dbdma_offsets[index][0];
+    props[3] = 0x0;
+    props[4] = dbdma_offsets[index][1];
+    props[5] = 0x0;
     NEWWORLD(set_property(dnode, "interrupts",
-             (char *)&props, 3 * sizeof(cell)));
+             (char *)&props, 6 * sizeof(cell)));
+
+    set_int_property(dnode, "slot-names", 0);
 
     device_end();
 
-    if (legacy) {
-        uart_init_line(
-                (unsigned char*)addr + IO_ESCC_LEGACY_OFFSET + offset * 0x4,
-                CONFIG_SERIAL_SPEED);
-    } else {
-        uart_init_line(
-                (unsigned char*)addr + IO_ESCC_OFFSET + offset * 0x20,
-                CONFIG_SERIAL_SPEED);
-    }
+    uart_init_line((unsigned char*)addr + offset + reg_offsets[legacy][index][0],
+                   CONFIG_SERIAL_SPEED);
 }
 
 void
@@ -494,6 +514,7 @@ escc_init(const char *path, phys_addr_t addr)
     set_property(dnode, "device_type", "escc",
                  strlen("escc") + 1);
     set_property(dnode, "compatible", "escc\0CHRP,es0", 14);
+    set_property(dnode, "ranges", "", 0);
 
     fword("finish-device");
 
@@ -521,6 +542,7 @@ escc_init(const char *path, phys_addr_t addr)
     set_property(dnode, "device_type", "escc-legacy",
                  strlen("escc-legacy") + 1);
     set_property(dnode, "compatible", "chrp,es1", 9);
+    set_property(dnode, "ranges", "", 0);
 
     fword("finish-device");
 
